local AceLocale = LibStub("AceLocale-3.0")
local L = AceLocale:GetLocale("Spy")

function Spy:RefreshCurrentList(player, source)
	local MainWindow = Spy.MainWindow
	if not MainWindow:IsShown() then
		return
	end

	local mode = Spy.db.profile.CurrentList
	local manageFunction = Spy.ListTypes[mode][2]
	if manageFunction then manageFunction() end

	local button = 1
	for index, data in pairs(Spy.CurrentList) do
		if button <= Spy.ButtonLimit then
			local level = "??"
			local class = "UNKNOWN"
			local opacity = 1

			local playerData = SpyPerCharDB.PlayerData[data.player]
			if playerData then
				if playerData.level then level = playerData.level end
				if playerData.class then class = playerData.class end
			end
			if mode == 1 and Spy.InactiveList[data.player] then
				opacity = 0.5
			end

			if player == data.player then
				if not source or source ~= UnitName("player") then
					Spy:AlertPlayer(player)
					if not source then Spy:AnnouncePlayer(player) end
				end
			end

			Spy:SetBar(button, data.player, level.." "..L[class], 100, "Class", class, nil, opacity)
			Spy.ButtonName[button] = data.player
			button = button + 1
		end
	end
	Spy.ListAmountDisplayed = button - 1

	if Spy.db.profile.ResizeSpy then
		Spy:AutomaticallyResize()
	end

	Spy:ManageBarsDisplayed()
end

function Spy:ManageNearbyList()
	local prioritiseKoS = Spy.db.profile.PrioritiseKoS

	local activeKoS = {}
	local active = {}
	for player in pairs(Spy.ActiveList) do
		if prioritiseKoS and SpyPerCharDB.KOSData[player] then
			table.insert(activeKoS, { player = player, time = Spy.NearbyList[player] })
		else
			table.insert(active, { player = player, time = Spy.NearbyList[player] })
		end
	end

	local inactiveKoS = {}
	local inactive = {}
	for player in pairs(Spy.InactiveList) do
		if prioritiseKoS and SpyPerCharDB.KOSData[player] then
			table.insert(inactiveKoS, { player = player, time = Spy.NearbyList[player] })
		else
			table.insert(inactive, { player = player, time = Spy.NearbyList[player] })
		end
	end

	table.sort(activeKoS, function(a, b) return a.time < b.time end)
	table.sort(inactiveKoS, function(a, b) return a.time < b.time end)
	table.sort(active, function(a, b) return a.time < b.time end)
	table.sort(inactive, function(a, b) return a.time < b.time end)

	local list = {}
	for player in pairs(activeKoS) do table.insert(list, activeKoS[player]) end
	for player in pairs(inactiveKoS) do table.insert(list, inactiveKoS[player]) end
	for player in pairs(active) do table.insert(list, active[player]) end
	for player in pairs(inactive) do table.insert(list, inactive[player]) end
	Spy.CurrentList = list
end

function Spy:ManageLastHourList()
	local list = {}
	for player in pairs(Spy.LastHourList) do
		table.insert(list, { player = player, time = Spy.LastHourList[player] })
	end
	table.sort(list, function(a, b) return a.time > b.time end)
	Spy.CurrentList = list
end

function Spy:ManageIgnoreList()
	local list = {}
	for player in pairs(SpyPerCharDB.IgnoreData) do
		table.insert(list, { player = player, time = SpyPerCharDB.PlayerData[player].time })
	end
	table.sort(list, function(a, b) return a.time > b.time end)
	Spy.CurrentList = list
end

function Spy:ManageKillOnSightList()
	local list = {}
	for player in pairs(SpyPerCharDB.KOSData) do
		table.insert(list, { player = player, time = SpyPerCharDB.PlayerData[player].time })
	end
	table.sort(list, function(a, b) return a.time > b.time end)
	Spy.CurrentList = list
end

function Spy:ManageExpirations()
	local mode = Spy.db.profile.CurrentList
	local expirationFunction = Spy.ListTypes[mode][3]
	if expirationFunction then expirationFunction() end
end

function Spy:ManageNearbyListExpirations()
	local expired = false
	local currentTime = time()
	for player in pairs(Spy.ActiveList) do
		if (currentTime - Spy.ActiveList[player]) > Spy.ActiveTimeout then
			Spy.InactiveList[player] = Spy.ActiveList[player]
			Spy.ActiveList[player] = nil
			expired = true
		end
	end
	if Spy.db.profile.RemoveUndetected ~= "Never" then
		for player in pairs(Spy.InactiveList) do
			if (currentTime - Spy.InactiveList[player]) > Spy.InactiveTimeout then
				Spy.InactiveList[player] = nil
				Spy.NearbyList[player] = nil
				expired = true
			end
		end
	end
	if expired then
		Spy:RefreshCurrentList()
		if Spy.db.profile.HideSpy then
			local entries = 0
			for _ in pairs(Spy.NearbyList) do
				entries = entries + 1
			end
			if entries == 0 then
				Spy.MainWindow:Hide()
			end
		end
	end
end

function Spy:ManageLastHourListExpirations()
	local expired = false
	local currentTime = time()
	for player in pairs(Spy.LastHourList) do
		if (currentTime - Spy.LastHourList[player]) > 3600 then
			Spy.LastHourList[player] = nil
			expired = true
		end
	end
	if expired then
		Spy:RefreshCurrentList()
	end
end

function Spy:RemovePlayerFromList(player)
	Spy.NearbyList[player] = nil
	Spy.ActiveList[player] = nil
	Spy.InactiveList[player] = nil
	Spy:RefreshCurrentList()
end

function Spy:ClearList()
	Spy.NearbyList = {}
	Spy.ActiveList = {}
	Spy.InactiveList = {}
	Spy.ListAmountDisplayed = 0
	Spy:SetCurrentList(1)
	if IsControlKeyDown() then
		Spy:EnableSpy(not Spy.db.profile.Enabled, false)
	end
end

function Spy:AddPlayerData(name, class, level, race, isEnemy, isGuess)
	local info = {}
	info.class = class
	info.level = level
	info.race = race
	info.isEnemy = isEnemy
	info.isGuess = isGuess
	info.time = time()
	if isEnemy then
		info.zone = GetZoneText()
		info.subZone = GetSubZoneText()
		info.mapX, info.mapY = GetPlayerMapPosition("player")
		info.mapX = math.floor(tonumber(info.mapX) * 100) / 100
		info.mapY = math.floor(tonumber(info.mapY) * 100) / 100
	end
	SpyPerCharDB.PlayerData[name] = info
end

function Spy:UpdatePlayerData(name, class, level, race, isEnemy, isGuess)
	local playerData = SpyPerCharDB.PlayerData[name]
	if not playerData then
		Spy:AddPlayerData(name, class, level, race, isEnemy, isGuess)
	else
		if class then SpyPerCharDB.PlayerData[name].class = class end
		if level then SpyPerCharDB.PlayerData[name].level = level end
		if race then SpyPerCharDB.PlayerData[name].race = race end
		if isEnemy then SpyPerCharDB.PlayerData[name].isEnemy = isEnemy end
		if isGuess then SpyPerCharDB.PlayerData[name].isGuess = isGuess end
		SpyPerCharDB.PlayerData[name].time = time()
		if SpyPerCharDB.PlayerData[name].isEnemy then
			SpyPerCharDB.PlayerData[name].zone = GetZoneText()
			SpyPerCharDB.PlayerData[name].subZone = GetSubZoneText()
			local mapX, mapY = GetPlayerMapPosition("player")
			mapX = math.floor(tonumber(mapX) * 100) / 100
			mapY = math.floor(tonumber(mapY) * 100) / 100
			SpyPerCharDB.PlayerData[name].mapX = mapX
			SpyPerCharDB.PlayerData[name].mapY = mapY
		end
	end
end

function Spy:RemovePlayerData(name)
	SpyPerCharDB.PlayerData[name] = nil
end

function Spy:AddIgnoreData(name)
	SpyPerCharDB.IgnoreData[name] = true
end

function Spy:RemoveIgnoreData(name)
	if SpyPerCharDB.IgnoreData[name] then
		SpyPerCharDB.IgnoreData[name] = nil
	end
end

function Spy:AddKOSData(name)
	SpyPerCharDB.KOSData[name] = time()
end

function Spy:RemoveKOSData(name)
	if SpyPerCharDB.KOSData[name] then
		SpyPerCharDB.KOSData[name] = nil
	end
end

function Spy:AlertPlayer(player)
	if SpyPerCharDB.KOSData[player] then
		if Spy.db.profile.EnableSound then
			PlaySoundFile("Interface\\AddOns\\Spy\\Sounds\\detected-kos.wav")
		end
		if Spy.db.profile.WarnOnKOS then
			local text = Spy.db.profile.Colors.Warning["Warning Text"]
			local msg = L["KoSWarning"]..player
			UIErrorsFrame:AddMessage(msg, text.r, text.g, text.b, 1.0, UIERRORS_HOLD_TIME)
		end
	else
		if Spy.db.profile.EnableSound then
			PlaySoundFile("Interface\\AddOns\\Spy\\Sounds\\detected-nearby.wav")
		end
	end
end

function Spy:AlertStealthPlayer(player)
	if Spy.db.profile.EnableSound then
		PlaySoundFile("Interface\\AddOns\\Spy\\Sounds\\detected-stealth.wav")
	end
	if Spy.db.profile.WarnOnStealth then
		local text = Spy.db.profile.Colors.Warning["Warning Text"]
		local msg = L["StealthWarning"]..player
		UIErrorsFrame:AddMessage(msg, text.r, text.g, text.b, 1.0, UIERRORS_HOLD_TIME)
	end
end

function Spy:AnnouncePlayer(player, channel)
	local msg = ""
	local isKOS = SpyPerCharDB.KOSData[player]
	local playerData = SpyPerCharDB.PlayerData[player]

	local announce = Spy.db.profile.Announce
	if announce == "Self" and not channel then
		if isKOS then
			msg = msg..L["SpySignatureColored"]..L["KillOnSightDetectedColored"]..player
		else
			msg = msg..L["SpySignatureColored"]..L["PlayerDetectedColored"]..player
		end
	else
		if isKOS then
			msg = msg..L["KillOnSightDetected"]..player
		else
			msg = msg..L["PlayerDetected"]..player
		end
	end
	if playerData then
		if playerData.zone then
			if playerData.subZone and playerData.subZone ~= "" and playerData.subZone ~= playerData.zone then
				msg = msg.." - "..playerData.subZone..", "..playerData.zone
			else
				msg = msg.." - "..playerData.zone
			end
		end
		if playerData.mapX and playerData.mapY then msg = msg.." ("..math.floor(tonumber(playerData.mapX) * 100)..","..math.floor(tonumber(playerData.mapY) * 100)..")" end
	end

	if channel then
		-- announce to selected channel
		SendChatMessage(msg, channel)
	else
		-- do not announce players on other realms
		if announce ~= "Self" and string.find(player, "-") then return end

		-- announce to standard channel
		if announce ~= "None" then
			if isKOS or not Spy.db.profile.OnlyAnnounceKoS then
				if announce == "Self" then
					DEFAULT_CHAT_FRAME:AddMessage(msg)
				elseif announce == "Party" then
					SendChatMessage(msg, "PARTY")
				elseif announce == "Guild" then
					SendChatMessage(msg, "GUILD")
				elseif announce == "Raid" then
					SendChatMessage(msg, "RAID")
				elseif announce == "LocalDefense" then
					SendChatMessage(msg, "CHANNEL", nil, GetChannelName(L["LocalDefenseChannelName"].." - "..GetZoneText()))
				end
			end
		end
	end

	-- announce to other Spy users
	if Spy.db.profile.ShareData then
		local class, level, race, zone, subZone, mapX, mapY = "", "", "", "", "", "", ""
		if playerData then
			if playerData.class then class = playerData.class end
			if playerData.level then level = playerData.level end
			if playerData.race then race = playerData.race end
			if playerData.zone then zone = playerData.zone end
			if playerData.subZone then subZone = playerData.subZone end
			if playerData.mapX then mapX = playerData.mapX end
			if playerData.mapY then mapY = playerData.mapY end
		end
		local details = Spy.Version.."|"..player.."|"..class.."|"..level.."|"..race.."|"..zone.."|"..subZone.."|"..mapX.."|"..mapY
		if strlen(details) < 240 then
			if channel then
				Spy:SendCommMessage(Spy.Signature, details, channel)
			else
				Spy:SendCommMessage(Spy.Signature, details, "PARTY")
				Spy:SendCommMessage(Spy.Signature, details, "RAID")
				Spy:SendCommMessage(Spy.Signature, details, "GUILD")
			end
		end
	end
end

function Spy:ToggleIgnorePlayer(ignore, player)
	if ignore then
		Spy:AddIgnoreData(player)
		Spy:RemoveKOSData(player)
		if Spy.db.profile.EnableSound then
			PlaySoundFile("Interface\\AddOns\\Spy\\Sounds\\list-add.wav")
		end
		DEFAULT_CHAT_FRAME:AddMessage(L["SpySignatureColored"]..L["PlayerAddedToIgnoreColored"]..player)
	else
		Spy:RemoveIgnoreData(player)
		if Spy.db.profile.EnableSound then
			PlaySoundFile("Interface\\AddOns\\Spy\\Sounds\\list-remove.wav")
		end
		DEFAULT_CHAT_FRAME:AddMessage(L["SpySignatureColored"]..L["PlayerRemovedFromIgnoreColored"]..player)
	end
	Spy:RefreshCurrentList()
end

function Spy:ToggleKOSPlayer(kos, player)
	if kos then
		Spy:AddKOSData(player)
		Spy:RemoveIgnoreData(player)
		if Spy.db.profile.EnableSound then
			PlaySoundFile("Interface\\AddOns\\Spy\\Sounds\\list-add.wav")
		end
		DEFAULT_CHAT_FRAME:AddMessage(L["SpySignatureColored"]..L["PlayerAddedToKOSColored"]..player)
	else
		Spy:RemoveKOSData(player)
		if Spy.db.profile.EnableSound then
			PlaySoundFile("Interface\\AddOns\\Spy\\Sounds\\list-remove.wav")
		end
		DEFAULT_CHAT_FRAME:AddMessage(L["SpySignatureColored"]..L["PlayerRemovedFromKOSColored"]..player)
	end
	Spy:RefreshCurrentList()
end

function Spy:ButtonClicked()
	local name = Spy.ButtonName[this.id]
	if name and name ~= "" then
		if arg1 == "LeftButton" then
			if IsShiftKeyDown() then
				if SpyPerCharDB.KOSData[name] then
					Spy:ToggleKOSPlayer(false, name)
				else
					Spy:ToggleKOSPlayer(true, name)
				end
			elseif IsControlKeyDown() then
				if SpyPerCharDB.IgnoreData[name] then
					Spy:ToggleIgnorePlayer(false, name)
				else
					Spy:ToggleIgnorePlayer(true, name)
				end
			else
				this:SetAttribute("macrotext", "/targetexact "..name)
			end
		elseif arg1 == "RightButton" then
			Spy:BarDropDownOpen(this)
			CloseDropDownMenus(1)
			ToggleDropDownMenu(1, nil, Spy_BarDropDownMenu)
		end
	end
end

function Spy:ParseUnitAbility(analyseSpell, event, guid, player, flags, spellId, spellName)
	local learnt = false
	if player then
		local class = nil
		local level = nil
		local race = nil
		local isEnemy = nil
		local isGuess = true
		local playerData = SpyPerCharDB.PlayerData[player]
		if bit.band(flags, COMBATLOG_OBJECT_REACTION_HOSTILE) == COMBATLOG_OBJECT_REACTION_HOSTILE then
			isEnemy = true
			if not playerData or playerData.isEnemy == nil then
				learnt = true
			end

			if analyseSpell then
				local type = strsub(event, 1, 5)
				if type == "SWING" or type == "SPELL" or type == "RANGE" then
					local ability = Spy_AbilityList[spellName]
					if ability then
						if ability.class and not (playerData and playerData.class) then
							class = ability.class
							learnt = true
						end
						if ability.level and (not playerData or not playerData.level or playerData.level < ability.level) then
							level = ability.level
							learnt = true
						end
						if ability.race and not (playerData and playerData.race) then
							race = ability.race
							learnt = true
						end
					end
					if class and race and level == Spy.MaximumPlayerLevel then
						isGuess = false
						learnt = true
					end
				end
			end
		else
			isEnemy = false
			isGuess = false
			learnt = true
		end

		Spy:UpdatePlayerData(player, class, level, race, isEnemy, isGuess)
		return learnt, SpyPerCharDB.PlayerData[player]
	end
	return learnt, nil
end

function Spy:ParseUnitDetails(player, class, level, race, zone, subZone, mapX, mapY)
	if player then
		local playerData = SpyPerCharDB.PlayerData[player]
		if not playerData then
			Spy:AddPlayerData(player, class, level, race, true, true)
		else
			if not playerData.class then SpyPerCharDB.PlayerData[player].class = class end
			if not playerData.level or playerData.level < level then SpyPerCharDB.PlayerData[player].level = level end
			if not playerData.race then SpyPerCharDB.PlayerData[player].race = race end
		end
		SpyPerCharDB.PlayerData[player].isEnemy = true
		SpyPerCharDB.PlayerData[player].time = time()
		SpyPerCharDB.PlayerData[player].zone = zone
		SpyPerCharDB.PlayerData[player].subZone = subZone
		SpyPerCharDB.PlayerData[player].mapX = mapX
		SpyPerCharDB.PlayerData[player].mapY = mapY

		return true, SpyPerCharDB.PlayerData[player]
	end
	return true, nil
end

function Spy:AddDetected(player, timestamp, learnt, source)
	if not Spy.NearbyList[player] then
		if Spy.db.profile.ShowOnDetection and not Spy.db.profile.MainWindowVis then
			Spy:SetCurrentList(1)
			Spy:EnableSpy(true, true)
		end
		if Spy.db.profile.CurrentList ~= 1 and Spy.db.profile.MainWindowVis and Spy.db.profile.ShowNearbyList then
			Spy:SetCurrentList(1)
		end

		if source and source ~= UnitName("player") and not Spy.ActiveList[player] then
			Spy.NearbyList[player] = timestamp
			Spy.LastHourList[player] = timestamp
			Spy.InactiveList[player] = timestamp
		else
			Spy.NearbyList[player] = timestamp
			Spy.LastHourList[player] = timestamp
			Spy.ActiveList[player] = timestamp
			Spy.InactiveList[player] = nil
		end

		if Spy.db.profile.CurrentList == 1 then
			Spy:RefreshCurrentList(player, source)
		else
			if not source or source ~= UnitName("player") then
				Spy:AlertPlayer(player)
				if not source then Spy:AnnouncePlayer(player) end
			end
		end
	elseif not Spy.ActiveList[player] then
		if Spy.db.profile.ShowOnDetection and not Spy.db.profile.MainWindowVis then
			Spy:SetCurrentList(1)
			Spy:EnableSpy(true, true)
		end
		if Spy.db.profile.CurrentList ~= 1 and Spy.db.profile.MainWindowVis and Spy.db.profile.ShowNearbyList then
			Spy:SetCurrentList(1)
		end

		Spy.LastHourList[player] = timestamp
		Spy.ActiveList[player] = timestamp
		Spy.InactiveList[player] = nil

		if Spy.db.profile.CurrentList == 1 then
			Spy:RefreshCurrentList()
		end
	else
		Spy.ActiveList[player] = timestamp
		if learnt and Spy.db.profile.CurrentList == 1 then
			Spy:RefreshCurrentList()
		end
	end
end

Spy.ListTypes = {
	{L["Nearby"], Spy.ManageNearbyList, Spy.ManageNearbyListExpirations},
	{L["LastHour"], Spy.ManageLastHourList, Spy.ManageLastHourListExpirations},
	{L["Ignore"], Spy.ManageIgnoreList},
	{L["KillOnSight"], Spy.ManageKillOnSightList},
}
